
#define NETBOOT_DIRECTORY			"/Library/NetBoot"
#define NETBOOT_SHAREPOINT_PREFIX		"NetBootSP"
#define NETBOOT_SHAREPOINT_LINK			".sharepoint"
#define NETBOOT_CLIENTS_SHAREPOINT_PREFIX	"NetBootClients"
#define NETBOOT_CLIENTS_SHAREPOINT_LINK		".clients"
#define NETBOOT_TFTP_DIRECTORY			"NetBoot"
#define NETBOOT_TFTP_PATH			"/private/tftpboot"

#define NETBOOT_IMAGE_SUFFIX		".nbi"
#define NETBOOT_IMAGE_INFO_PLIST	"NBImageInfo.plist"

#define NETBOOT_GROUP			"netboot"
#define NETBOOT_USER_PREFIX		"netboot"
