/*
 * Copyright (c) 1999-2003 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * This file was auto-generated by ./genoptionfiles -dhcptype, do not edit!
 */
#ifndef _S_DHCP_TYPE
#define _S_DHCP_TYPE

typedef enum {
    dhcptype_first_e    	 = 0,
    dhcptype_none_e     	 = dhcptype_first_e,
    dhcptype_opaque_e   	 = 1,
    dhcptype_bool_e     	 = 2,
    dhcptype_uint8_e    	 = 3,
    dhcptype_uint16_e   	 = 4,
    dhcptype_uint32_e   	 = 5,
    dhcptype_int32_e    	 = 6,
    dhcptype_uint8_mult_e	 = 7,
    dhcptype_uint16_mult_e	 = 8,
    dhcptype_string_e   	 = 9,
    dhcptype_ip_e       	 = 10,
    dhcptype_ip_mult_e  	 = 11,
    dhcptype_ip_pairs_e 	 = 12,
    dhcptype_last_e     	 = dhcptype_ip_pairs_e,
} dhcptype_t;

typedef struct {
    dhcptype_t	type;
    unsigned char *	name;
} dhcptag_info_t;

typedef struct {
    int		size;  /* in bytes */
    int		multiple_of; /* type of element */
    unsigned char * name;
} dhcptype_info_t;

#endif _S_DHCP_TYPE
